PHARMSPEC_SOFTWARE_NAME = "PharmSpec"

# This map is used to map the column names to units, for calculated data items.
UNIT_LOOKUP = {
    "particle_size": "μm",
    "cumulative_count": "(unitless)",
    "cumulative_particle_density": "Counts/mL",
    "differential_particle_density": "Counts/mL",
    "differential_count": "(unitless)",
}
VALID_CALCS = ["Average"]
